from packaging import version

from django.conf import settings
from extras.plugins import PluginMenuItem, PluginMenuButton
from utilities.choices import ButtonColorChoices

# compatibility with netbox v3.3 that does not have PluginMenu
try:
    from extras.plugins import PluginMenu
    HAVE_MENU = True
except ImportError:
    HAVE_MENU = False
    PluginMenu = PluginMenuItem


asset_buttons = [
    PluginMenuButton(
        link='plugins:netbox_inventory:asset_add',
        title='Add',
        icon_class='mdi mdi-plus-thick',
        color=ButtonColorChoices.GREEN
    ),
    PluginMenuButton(
        link='plugins:netbox_inventory:asset_import',
        title='Import',
        icon_class='mdi mdi-upload',
        color=ButtonColorChoices.CYAN,
    )
]

supplier_buttons = [
    PluginMenuButton(
        link='plugins:netbox_inventory:supplier_add',
        title='Add',
        icon_class='mdi mdi-plus-thick',
        color=ButtonColorChoices.GREEN
    ),
    PluginMenuButton(
        link='plugins:netbox_inventory:supplier_import',
        title='Import',
        icon_class='mdi mdi-upload',
        color=ButtonColorChoices.CYAN,
    )
]

purchase_buttons = [
    PluginMenuButton(
        link='plugins:netbox_inventory:purchase_add',
        title='Add',
        icon_class='mdi mdi-plus-thick',
        color=ButtonColorChoices.GREEN
    ),
    PluginMenuButton(
        link='plugins:netbox_inventory:purchase_import',
        title='Import',
        icon_class='mdi mdi-upload',
        color=ButtonColorChoices.CYAN,
    )
]

inventoryitemtype_buttons = [
    PluginMenuButton(
        link='plugins:netbox_inventory:inventoryitemtype_add',
        title='Add',
        icon_class='mdi mdi-plus-thick',
        color=ButtonColorChoices.GREEN
    ),
    PluginMenuButton(
        link='plugins:netbox_inventory:inventoryitemtype_import',
        title='Import',
        icon_class='mdi mdi-upload',
        color=ButtonColorChoices.CYAN,
    )
]

inventoryitemgroup_buttons = [
    PluginMenuButton(
        link='plugins:netbox_inventory:inventoryitemgroup_add',
        title='Add',
        icon_class='mdi mdi-plus-thick',
        color=ButtonColorChoices.GREEN
    ),
    PluginMenuButton(
        link='plugins:netbox_inventory:inventoryitemgroup_import',
        title='Import',
        icon_class='mdi mdi-upload',
        color=ButtonColorChoices.CYAN,
    )
]

menu_buttons = (
    PluginMenuItem(
        link='plugins:netbox_inventory:asset_list',
        link_text='Assets',
        permissions=["netbox_inventory.view_asset"],
        buttons=asset_buttons,
    ),
    PluginMenuItem(
        link='plugins:netbox_inventory:supplier_list',
        link_text='Suppliers',
        permissions=["netbox_inventory.view_supplier"],
        buttons=supplier_buttons,
    ),
    PluginMenuItem(
        link='plugins:netbox_inventory:purchase_list',
        link_text='Purchases',
        permissions=["netbox_inventory.view_purchase"],
        buttons=purchase_buttons,
    ),
    PluginMenuItem(
        link='plugins:netbox_inventory:inventoryitemtype_list',
        link_text='Inventory Item Types',
        permissions=["netbox_inventory.view_inventoryitemtype"],
        buttons=inventoryitemtype_buttons,
    ),
    PluginMenuItem(
        link='plugins:netbox_inventory:inventoryitemgroup_list',
        link_text='Inventory Item Groups',
        permissions=["netbox_inventory.view_inventoryitemgroup"],
        buttons=inventoryitemgroup_buttons,
    ),
)

# can't use utils.get_plugin_setting() here, get value manually
if (HAVE_MENU and settings.PLUGINS_CONFIG['netbox_inventory']['top_level_menu']):
    # add a top level entry
    menu = PluginMenu(
        label=f'Inventory',
        groups=(
            ('Asset Management', menu_buttons),
        ),
        icon_class='mdi mdi-clipboard-text-multiple-outline'
    )
else:
    # display under plugins
    menu_items = menu_buttons



from netbox.views import generic
from utilities.utils import count_related
from .. import filtersets, forms, models, tables

__all__ = (
    'PurchaseView',
    'PurchaseListView',
    'PurchaseEditView',
    'PurchaseDeleteView',
    'PurchaseBulkImportView',
    'PurchaseBulkEditView',
    'PurchaseBulkDeleteView',
)

class PurchaseView(generic.ObjectView):
    queryset = models.Purchase.objects.all()

    def get_extra_context(self, request, instance):
        purchase_assets = models.Asset.objects.restrict(request.user, 'view').filter(
            purchase=instance
        )
        asset_table = tables.AssetTable(purchase_assets, user=request.user)
        asset_table.columns.hide('purchase')
        asset_table.columns.hide('purchase_date')
        asset_table.columns.hide('supplier')
        asset_table.configure(request)

        return {
            'asset_table': asset_table,
            'asset_count': models.Asset.objects.filter(purchase=instance).count(),
        }

class PurchaseListView(generic.ObjectListView):
    queryset = models.Purchase.objects.annotate(
        asset_count=count_related(models.Asset, 'purchase'),
    )
    table = tables.PurchaseTable
    filterset = filtersets.PurchaseFilterSet
    filterset_form = forms.PurchaseFilterForm


class PurchaseEditView(generic.ObjectEditView):
    queryset = models.Purchase.objects.all()
    form = forms.PurchaseForm


class PurchaseDeleteView(generic.ObjectDeleteView):
    queryset = models.Purchase.objects.all()


class PurchaseBulkImportView(generic.BulkImportView):
    queryset = models.Purchase.objects.all()   
    table = tables.PurchaseTable 
    model_form = forms.PurchaseImportForm


class PurchaseBulkEditView(generic.BulkEditView):
    queryset = models.Purchase.objects.all()
    filterset = filtersets.PurchaseFilterSet
    table = tables.PurchaseTable
    form = forms.PurchaseBulkEditForm


class PurchaseBulkDeleteView(generic.BulkDeleteView):
    queryset = models.Purchase.objects.all()
    table = tables.PurchaseTable



from .asset import *
from .inventoryitem_type import *
from .inventoryitem_group import *
from .supplier import *
from .purchase import *
from .asset_create import *
from .asset_assign import *
from .tabs import *



from django.urls import path

from netbox.views.generic import ObjectChangeLogView, ObjectJournalView
from . import models, views


urlpatterns = (

    # Assets
    path('assets/', views.AssetListView.as_view(), name='asset_list'),
    path('assets/add/', views.AssetEditView.as_view(), name='asset_add'),
    path('assets/bulk-add/', views.AssetBulkCreateView.as_view(), name='asset_bulk_add'),
    path('assets/import/', views.AssetBulkImportView.as_view(), name='asset_import'),
    path('assets/edit/', views.AssetBulkEditView.as_view(), name='asset_bulk_edit'),
    path('assets/delete/', views.AssetBulkDeleteView.as_view(), name='asset_bulk_delete'),
    path('assets/<int:pk>', views.AssetView.as_view(), name='asset'),
    path('assets/<int:pk>/edit/', views.AssetEditView.as_view(), name='asset_edit'),
    path('assets/<int:pk>/delete/', views.AssetDeleteView.as_view(), name='asset_delete'),
    path('assets/<int:pk>/changelog/', ObjectChangeLogView.as_view(), name='asset_changelog', kwargs={'model': models.Asset}),
    path('assets/<int:pk>/journal/', ObjectJournalView.as_view(), name='asset_journal', kwargs={'model': models.Asset}),
    path('assets/<int:pk>/assign/', views.AssetAssignView.as_view(), name='asset_assign'),
    path('assets/device/create/', views.AssetDeviceCreateView.as_view(), name='asset_device_create'),
    path('assets/module/create/', views.AssetModuleCreateView.as_view(), name='asset_module_create'),
    path('assets/inventory-item/create/', views.AssetInventoryItemCreateView.as_view(), name='asset_inventoryitem_create'),

    # Suppliers
    path('suppliers/', views.SupplierListView.as_view(), name='supplier_list'),
    path('suppliers/add/', views.SupplierEditView.as_view(), name='supplier_add'),
    path('suppliers/import/', views.SupplierBulkImportView.as_view(), name='supplier_import'),
    path('suppliers/edit/', views.SupplierBulkEditView.as_view(), name='supplier_bulk_edit'),
    path('suppliers/delete/', views.SupplierBulkDeleteView.as_view(), name='supplier_bulk_delete'),
    path('suppliers/<int:pk>', views.SupplierView.as_view(), name='supplier'),
    path('suppliers/<int:pk>/edit/', views.SupplierEditView.as_view(), name='supplier_edit'),
    path('suppliers/<int:pk>/delete/', views.SupplierDeleteView.as_view(), name='supplier_delete'),
    path('suppliers/<int:pk>/changelog/', ObjectChangeLogView.as_view(), name='supplier_changelog', kwargs={'model': models.Supplier}),
    path('suppliers/<int:pk>/journal/', ObjectJournalView.as_view(), name='supplier_journal', kwargs={'model': models.Supplier}),

    # Purchases
    path('purchases/', views.PurchaseListView.as_view(), name='purchase_list'),
    path('purchases/add/', views.PurchaseEditView.as_view(), name='purchase_add'),
    path('purchases/import/', views.PurchaseBulkImportView.as_view(), name='purchase_import'),
    path('purchases/edit/', views.PurchaseBulkEditView.as_view(), name='purchase_bulk_edit'),
    path('purchases/delete/', views.PurchaseBulkDeleteView.as_view(), name='purchase_bulk_delete'),
    path('purchases/<int:pk>', views.PurchaseView.as_view(), name='purchase'),
    path('purchases/<int:pk>/edit/', views.PurchaseEditView.as_view(), name='purchase_edit'),
    path('purchases/<int:pk>/delete/', views.PurchaseDeleteView.as_view(), name='purchase_delete'),
    path('purchases/<int:pk>/changelog/', ObjectChangeLogView.as_view(), name='purchase_changelog', kwargs={'model': models.Purchase}),
    path('purchases/<int:pk>/journal/', ObjectJournalView.as_view(), name='purchase_journal', kwargs={'model': models.Purchase}),

    # InventoryItemTypes
    path('inventory-item-types/', views.InventoryItemTypeListView.as_view(), name='inventoryitemtype_list'),
    path('inventory-item-types/add/', views.InventoryItemTypeEditView.as_view(), name='inventoryitemtype_add'),
    path('inventory-item-types/import/', views.InventoryItemTypeBulkImportView.as_view(), name='inventoryitemtype_import'),
    path('inventory-item-types/edit/', views.InventoryItemTypeBulkEditView.as_view(), name='inventoryitemtype_bulk_edit'),
    path('inventory-item-types/delete/', views.InventoryItemTypeBulkDeleteView.as_view(), name='inventoryitemtype_bulk_delete'),
    path('inventory-item-types/<int:pk>', views.InventoryItemTypeView.as_view(), name='inventoryitemtype'),
    path('inventory-item-types/<int:pk>/edit/', views.InventoryItemTypeEditView.as_view(), name='inventoryitemtype_edit'),
    path('inventory-item-types/<int:pk>/delete/', views.InventoryItemTypeDeleteView.as_view(), name='inventoryitemtype_delete'),
    path('inventory-item-types/<int:pk>/changelog/', ObjectChangeLogView.as_view(), name='inventoryitemtype_changelog', kwargs={'model': models.InventoryItemType}),

    # InventoryItemGroups
    path('inventory-item-groups/', views.InventoryItemGroupListView.as_view(), name='inventoryitemgroup_list'),
    path('inventory-item-groups/add/', views.InventoryItemGroupEditView.as_view(), name='inventoryitemgroup_add'),
    path('inventory-item-groups/import/', views.InventoryItemGroupBulkImportView.as_view(), name='inventoryitemgroup_import'),
    path('inventory-item-groups/edit/', views.InventoryItemGroupBulkEditView.as_view(), name='inventoryitemgroup_bulk_edit'),
    path('inventory-item-groups/delete/', views.InventoryItemGroupBulkDeleteView.as_view(), name='inventoryitemgroup_bulk_delete'),
    path('inventory-item-groups/<int:pk>', views.InventoryItemGroupView.as_view(), name='inventoryitemgroup'),
    path('inventory-item-groups/<int:pk>/edit/', views.InventoryItemGroupEditView.as_view(), name='inventoryitemgroup_edit'),
    path('inventory-item-groups/<int:pk>/delete/', views.InventoryItemGroupDeleteView.as_view(), name='inventoryitemgroup_delete'),
    path('inventory-item-groups/<int:pk>/changelog/', ObjectChangeLogView.as_view(), name='inventoryitemgroup_changelog', kwargs={'model': models.InventoryItemGroup}),

)



from rest_framework import serializers

from dcim.api.serializers import (
    NestedDeviceTypeSerializer, NestedDeviceSerializer,
    NestedManufacturerSerializer,
    NestedModuleTypeSerializer, NestedModuleSerializer,
    NestedInventoryItemSerializer
)
from tenancy.api.serializers import NestedContactSerializer, NestedTenantSerializer
from netbox.api.serializers import NetBoxModelSerializer
from .nested_serializers import *
from ..models import Asset, InventoryItemType, InventoryItemGroup, Purchase, Supplier


class AssetSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='plugins-api:netbox_inventory-api:asset-detail'
    )
    device_type = NestedDeviceTypeSerializer(required=False, allow_null=True, default=None)
    device = NestedDeviceSerializer(required=False, allow_null=True, default=None)
    module_type = NestedModuleTypeSerializer(required=False, allow_null=True, default=None)
    module = NestedModuleSerializer(required=False, allow_null=True, default=None)
    inventoryitem_type = NestedInventoryItemTypeSerializer(required=False, allow_null=True, default=None) 
    inventoryitem = NestedInventoryItemSerializer(required=False, allow_null=True, default=None)
    storage_location = NestedModuleSerializer(required=False, allow_null=True, default=None)
    purchase = NestedPurchaseSerializer(required=False, allow_null=True, default=None)
    tenant = NestedTenantSerializer(required=False, allow_null=True, default=None)
    contact = NestedContactSerializer(required=False, allow_null=True, default=None)
    owner = NestedTenantSerializer(required=False, allow_null=True, default=None)


    class Meta:
        model = Asset
        fields = (
            'id', 'url', 'display', 'name', 'asset_tag', 'serial', 'status',
            'kind', 'device_type', 'device', 'module_type', 'module', 'inventoryitem_type','inventoryitem', 
            'tenant', 'contact', 'storage_location', 'owner', 'purchase',
            'warranty_start', 'warranty_end',
            'comments', 'tags', 'custom_fields', 'created', 'last_updated'
        )
        # DRF autiomatically creates validator from model's unique_together contraints
        # that doesn't work if we allow some filelds in a unique_together to be null
        # so we remove DRF's auto generated validators and rely on model's validation
        # logic to handle validation
        # see  https://www.django-rest-framework.org/api-guide/validators/#optional-fields
        validators = []


class SupplierSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='plugins-api:netbox_inventory-api:supplier-detail'
    )
    asset_count = serializers.IntegerField(read_only=True)
    purchase_count = serializers.IntegerField(read_only=True)
    
    class Meta:
        model = Supplier
        fields = (
            'id', 'url', 'display', 'name', 'slug', 'description', 'comments',
            'tags', 'custom_fields', 'created', 'last_updated', 'asset_count',
            'purchase_count',
        )


class PurchaseSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='plugins-api:netbox_inventory-api:purchase-detail'
    )
    supplier = NestedSupplierSerializer()
    asset_count = serializers.IntegerField(read_only=True)
    
    class Meta:
        model = Purchase
        fields = (
            'id', 'url', 'display', 'supplier', 'name', 'date', 'description', 'comments',
            'tags', 'custom_fields', 'created', 'last_updated', 'asset_count',
        )


class InventoryItemTypeSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='plugins-api:netbox_inventory-api:inventoryitemtype-detail'
    )
    manufacturer = NestedManufacturerSerializer()
    inventoryitem_group = NestedInventoryItemGroupSerializer(required=False, allow_null=True, default=None)
    asset_count = serializers.IntegerField(read_only=True)
    
    class Meta:
        model = InventoryItemType
        fields = (
            'id', 'url', 'display', 'model', 'slug', 'manufacturer', 'part_number',
            'inventoryitem_group', 'comments', 'tags', 'custom_fields', 'created',
            'last_updated', 'asset_count',
        )


class InventoryItemGroupSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='plugins-api:netbox_inventory-api:inventoryitemgroup-detail'
    )
    parent = NestedInventoryItemGroupSerializer(required=False, allow_null=True, default=None)
    asset_count = serializers.IntegerField(read_only=True)
    
    class Meta:
        model = InventoryItemGroup
        fields = (
            'id', 'url', 'display', 'name', 'parent', 'comments', 'tags', 'custom_fields',
            'created', 'last_updated', 'asset_count',
        )



from django.contrib.contenttypes.models import ContentType
from dcim.models import Device, DeviceRole, DeviceType, InventoryItem, Manufacturer, ModuleType, Site
from users.models import ObjectPermission
from utilities.testing import APIViewTestCases, disable_warnings
from rest_framework import status

from ..custom import APITestCase
from ...models import Asset, InventoryItemType


class AssetTest(
        APITestCase, 
        APIViewTestCases.GetObjectViewTestCase,
        APIViewTestCases.ListObjectsViewTestCase,
        APIViewTestCases.CreateObjectViewTestCase,
        APIViewTestCases.UpdateObjectViewTestCase,
        APIViewTestCases.DeleteObjectViewTestCase):
    model = Asset
    brief_fields = ['display', 'id', 'serial', 'url']

    bulk_update_data = {
        'status': 'used',
    }

    def test_assign_device_matching_device_type(self):
        """
        check assigning device to asset when asset's & device's device_type matches
        """
        # Add object-level permission
        obj_perm = ObjectPermission(
            name='Test permission',
            actions=['add', 'change']
        )
        obj_perm.save()
        obj_perm.users.add(self.user)
        obj_perm.object_types.add(ContentType.objects.get_for_model(self.model))

        update_data = {'device':self.device1.pk}

        response = self.client.post(self._get_list_url(), self.create_data[0], format='json', **self.header)
        instance = self._get_queryset().get(pk=response.data['id'])
        url = self._get_detail_url(instance)
        response = self.client.patch(url, update_data, format='json', **self.header)
        self.assertHttpStatus(response, status.HTTP_200_OK)
        instance.refresh_from_db()
        self.assertInstanceEqual(
            instance,
            update_data,
            exclude=self.validation_excluded_fields,
            api=True
        )

    def test_assign_device_missmatch_device_type(self):
        """
        check assigning device to asset when asset's & device's device_type doesn't match
        """
        # Add object-level permission
        obj_perm = ObjectPermission(
            name='Test permission',
            actions=['add', 'change']
        )
        obj_perm.save()
        obj_perm.users.add(self.user)
        obj_perm.object_types.add(ContentType.objects.get_for_model(self.model))

        update_data = {'device':self.device2.pk}

        response = self.client.post(self._get_list_url(), self.create_data[0], format='json', **self.header)
        instance = self._get_queryset().get(pk=response.data['id'])
        url = self._get_detail_url(instance)
        response = self.client.patch(url, update_data, format='json', **self.header)
        with disable_warnings('django.request'):
            self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)

    def test_assign_inventoryitem_to_asset_device(self):
        """
        check assigning inventoryitem to asset when asset.kind is device
        """
        # Add object-level permission
        obj_perm = ObjectPermission(
            name='Test permission',
            actions=['add', 'change']
        )
        obj_perm.save()
        obj_perm.users.add(self.user)
        obj_perm.object_types.add(ContentType.objects.get_for_model(self.model))

        update_data = {'inventoryitem':self.inventoryitem1.pk}

        response = self.client.post(self._get_list_url(), self.create_data[0], format='json', **self.header)
        instance = self._get_queryset().get(pk=response.data['id'])
        url = self._get_detail_url(instance)
        response = self.client.patch(url, update_data, format='json', **self.header)
        with disable_warnings('django.request'):
            self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)

    @classmethod
    def setUpTestData(cls):
        manufacturer = Manufacturer.objects.create(name='Manufacturer 1', slug='manufacturer1')
        device_type1 = DeviceType.objects.create(model='Device Type 1', slug='devicetype1', manufacturer=manufacturer)
        device_type2 = DeviceType.objects.create(model='Device Type 2', slug='devicetype2', manufacturer=manufacturer)
        module_type1 = ModuleType.objects.create(model='Module Type 1', manufacturer=manufacturer)
        inventoryitem_type1 = InventoryItemType.objects.create(model='II Type 1', manufacturer=manufacturer)
        site1 = Site.objects.create(name='Site 1', slug='site1')
        device_role1 = DeviceRole.objects.create(name='Device Role 1', slug='devicerole1')
        cls.device1 = Device.objects.create(name='Device 1', device_role=device_role1, device_type=device_type1, site=site1, status='active')
        cls.device2 = Device.objects.create(name='Device 2', device_role=device_role1, device_type=device_type2, site=site1, status='active')
        cls.inventoryitem1 = InventoryItem.objects.create(device=cls.device1, name='II 1')

        Asset.objects.create(name='Asset 1', serial='asset1', device_type=device_type1)
        Asset.objects.create(name='Asset 2', serial='asset2', device_type=device_type1)
        Asset.objects.create(name='Asset 3', serial='asset3', device_type=device_type1)

        cls.create_data = [
            {
                'name': 'Asset 4',
                'serial': 'asset4',
                'status': 'stored',
                'device_type': device_type1.pk,
                'device': None,
            },
            {
                'name': 'Asset 5',
                'serial': 'asset5',
                'status': 'stored',
                'device_type': None,
                'device': None,
                'module_type': module_type1.pk,
                'module': None,
            },
            {
                'name': 'Asset 6',
                'serial': 'asset6',
                'status': 'stored',
                'inventoryitem_type': inventoryitem_type1.pk,
                'inventoryitem': cls.inventoryitem1.pk,
            },
        ]



import datetime
from utilities.testing import ViewTestCases

from netbox_inventory.tests.custom import ModelViewTestCase
from netbox_inventory.models import Purchase, Supplier


class PurchaseTestCase(
    ModelViewTestCase,
    ViewTestCases.PrimaryObjectViewTestCase,
):

    model = Purchase

    @classmethod
    def setUpTestData(cls):
        supplier1 = Supplier.objects.create(
            name='Supplier 1',
            slug='supplier1',
        )
        supplier2 = Supplier.objects.create(
            name='Supplier 2',
            slug='supplier2',
        )
        purchase1 = Purchase.objects.create(
            name='Purchase 1',
            supplier=supplier1,
        )
        purchase2 = Purchase.objects.create(
            name='Purchase 2',
            supplier=supplier1,
        )
        purchase3 = Purchase.objects.create(
            name='Purchase 3',
            supplier=supplier1,
        )
        cls.form_data = {
            'name': 'Purchase',
            'supplier': supplier1.pk,
            'description': 'Purchase description',
            'date': datetime.date(day=1, month=1, year=2023),
        }
        cls.csv_data = (
            'name,supplier,date',
            f'Purchase 4,{supplier1.name},2023-03-26',
            f'Purchase 5,{supplier1.name},2023-03-26',
            f'Purchase 6,{supplier1.name},2023-03-26',
        )
        cls.csv_update_data = (
            'id,description,supplier',
            f'{purchase1.pk},description 1,{supplier2.name}',
            f'{purchase2.pk},description 2,{supplier2.name}',
            f'{purchase3.pk},description 3,{supplier2.name}',
        )
        cls.bulk_edit_data = {
            'description': 'bulk description',
            'date': datetime.date(day=1, month=1, year=2022),
        }



from dcim.models import Site, Location, Rack
from netbox.views import generic
from utilities.views import ViewTab, register_model_view

from ..models import Asset
from ..tables import AssetTable
from ..filtersets import AssetFilterSet
from ..utils import query_located


class BaseAssetsTab(generic.ObjectChildrenView):
    """
    Shows Assets tab with table of assets located here
    Can be filtered on installed or stored assets or both.
    Accepts query param assets_shown:
    - stored - show currently stored here (storage_location & stored status)
    - installed - show currently in use here (installed as a or into a device)
    - all - show both
    """
    child_model = Asset
    table = AssetTable
    filterset = AssetFilterSet
    tab = ViewTab(
        label='Assets',
        permission='netbox_inventory.view_asset'
    )

    def get_children(self, request, parent):
        """ Returns queryset that populates the table """
        assets_shown = request.GET.get('assets_shown', 'all')
        typ = parent._meta.model.__name__.lower()
        return query_located(
            Asset.objects.all(),
            typ,
            [parent.pk],
            assets_shown
        )

    def get_extra_context(self, request, instance):
        context = super().get_extra_context(request, instance)
        context['assets_shown'] = request.GET.get('assets_shown', 'all')
        return context


@register_model_view(Site, name='inventory_assets', path='assets')
class SiteAssets(BaseAssetsTab):
    queryset = Site.objects.all()
    template_name = "netbox_inventory/tabs/located_assets_site.html"


@register_model_view(Location, name='inventory_assets', path='assets')
class LocationAssets(BaseAssetsTab):
    queryset = Location.objects.all()
    template_name = "netbox_inventory/tabs/located_assets_location.html"

@register_model_view(Rack, name='inventory_assets', path='assets')
class LocationAssets(BaseAssetsTab):
    queryset = Rack.objects.all()
    template_name = "netbox_inventory/tabs/located_assets_rack.html"



import logging

from django.dispatch import receiver
from django.db.models.signals import pre_save, pre_delete

from dcim.models import Device, Module, InventoryItem
from utilities.exceptions import AbortRequest
from .models import Asset
from .utils import get_plugin_setting, get_status_for


logger = logging.getLogger('netbox.netbox_inventory.signals')


@receiver(pre_save, sender=Device)
@receiver(pre_save, sender=Module)
@receiver(pre_save, sender=InventoryItem)
def prevent_update_serial_asset_tag(instance, **kwargs):
    """
    When a hardware (Device, Module or InventoryItem) has an Asset assigned and
    user changes serial or asset_tag on hardware, prevent that change
    and inform that change must be made on Asset instance instead.
    
    Only enforces if `sync_hardware_serial_asset_tag` setting is true.
    """
    try:
        # will raise RelatedObjectDoesNotExist if not set
        asset = instance.assigned_asset
    except Asset.DoesNotExist:
        return
    if not get_plugin_setting('sync_hardware_serial_asset_tag'):
        # don't enforce if sync not enabled
        return
    if asset.serial != instance.serial or asset.asset_tag != instance.asset_tag:
        raise AbortRequest(f'Cannot change {asset.kind} serial and asset tag if asset is assigned. Please update via inventory > asset instead.')


@receiver(pre_delete, sender=Device)
@receiver(pre_delete, sender=Module)
@receiver(pre_delete, sender=InventoryItem)
def free_assigned_asset(instance, **kwargs):
    """
    If a hardware (Device, Module or InventoryItem) has an Asset assigned and
    that hardware is deleted, update Asset.status to stored_status.

    Netbox handles deletion in a DB transaction, so if deletion failes for any
    reason, this status change will also be reverted.
    """
    stored_status = get_status_for('stored')
    if not stored_status:
        return
    try:
        # will raise RelatedObjectDoesNotExist if not set
        asset = instance.assigned_asset
    except Asset.DoesNotExist:
        return
    asset.snapshot()
    asset.status = stored_status
    # also unassign that item from asset
    setattr(asset, asset.kind, None)
    asset.full_clean()
    asset.save()
    logger.info(f'Asset marked as stored {asset}')



# Generated by Django 4.1.7 on 2023-03-26 14:56

from django.db import migrations, models
import django.db.models.deletion
import mptt.fields


def rebuild_tree(apps, schema_editor):
    from ..models import InventoryItemGroup 
    if hasattr(InventoryItemGroup,'_tree_manager'):
        InventoryItemGroup._tree_manager.rebuild()


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_inventory', '0003_add_inventoryitemgroup'),
    ]

    operations = [
        migrations.AddField(
            model_name='inventoryitemgroup',
            name='description',
            field=models.CharField(blank=True, max_length=200),
        ),
        migrations.AddField(
            model_name='inventoryitemgroup',
            name='level',
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='inventoryitemgroup',
            name='lft',
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='inventoryitemgroup',
            name='parent',
            field=mptt.fields.TreeForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='netbox_inventory.inventoryitemgroup'),
        ),
        migrations.AddField(
            model_name='inventoryitemgroup',
            name='rght',
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='inventoryitemgroup',
            name='tree_id',
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='inventoryitemgroup',
            name='name',
            field=models.CharField(max_length=100),
        ),
        migrations.RunPython(rebuild_tree, migrations.RunPython.noop),
        migrations.AddConstraint(
            model_name='inventoryitemgroup',
            constraint=models.UniqueConstraint(fields=('parent', 'name'), name='netbox_inventory_inventoryitemgroup_parent_name'),
        ),
        migrations.AddConstraint(
            model_name='inventoryitemgroup',
            constraint=models.UniqueConstraint(condition=models.Q(('parent__isnull', True)), fields=('name',), name='netbox_inventory_inventoryitemgroup_name', violation_error_message='A top-level group with this name already exists.'),
        ),
    ]



from django import forms

from dcim.models import Device, InventoryItem, Module, Site
from netbox.forms import NetBoxModelForm
from utilities.forms import APISelect, DynamicModelChoiceField
from ..models import Asset

__all__ = (
    'AssetDeviceAssignForm',
    'AssetModuleAssignForm',
    'AssetInventoryItemAssignForm',
)


class AssetAssignMixin(forms.Form):
    name = forms.CharField(
        required=False,
        label='Asset name',
        help_text=Asset._meta.get_field('name').help_text
    )
    tags = None

    def _clean_hardware_type(self, kind):
        hardware_type = self.cleaned_data[f'{kind}_type']
        if getattr(self.instance, f'{kind}_type') != hardware_type:
            raise forms.ValidationError(f'You are not allowed to change {kind}_type')
        return hardware_type

    def _clean_hardware(self, kind):
        """
        Args: 
            kind (str): one of device, module or inventoryitem
        """
        hardware = self.cleaned_data[kind]
        if hardware:
            instance_hardware = getattr(self.instance, kind)
            if instance_hardware == hardware:
                # field was not changed
                return hardware

            if getattr(hardware, f'{kind}_type') != getattr(self.instance, f'{kind}_type'):
                raise forms.ValidationError(f'{kind} type of selected {kind} does not match asset\'s {kind} type')

            if Asset.objects.filter(**{kind: hardware}).exists():
                raise forms.ValidationError(f'{kind} {hardware} already has asset assigned')

        setattr(self.instance, kind, hardware)
        return hardware


class AssetDeviceAssignForm(AssetAssignMixin, NetBoxModelForm):
    site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        required=False,
        initial_params={'devices': '$device'},
    )
    device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        query_params={'device_type_id': '$device_type', 'site_id': '$site'},
        label='Device',
        required=False,
        help_text='Set to empty to unassign asset from device',
        widget=APISelect(
            api_url='/api/plugins/inventory/dcim/devices/',
            attrs={
                'data-static-params': '[{"queryParam":"has_asset_assigned","queryValue":"false"}]',
            },
        )
    )

    fieldsets = (
        ('Asset', ('device_type', 'name')),
        ('Device', ('site', 'device')),
        ('Tenancy', ('tenant', 'contact')),
    )

    class Meta:
        model = Asset
        fields = ('device_type', 'name', 'site', 'device', 'tenant', 'contact')
        widgets = {'device_type': forms.HiddenInput()}

    def clean_device_type(self):
        return self._clean_hardware_type('device')

    def clean_device(self):
        return self._clean_hardware('device')


class AssetModuleAssignForm(AssetAssignMixin, NetBoxModelForm):
    site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        required=False,
        initial_params={'devices': '$device'},
    )
    device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        query_params={'site_id': '$site'},
        label='Device',
        required=False,
    )
    module = DynamicModelChoiceField(
        queryset=Module.objects.all(),
        query_params={'module_type_id': '$module_type', 'device_id': '$device'},
        label='Module',
        required=False,
        help_text='Set to empty to unassign asset from module',
        widget=APISelect(
            api_url='/api/plugins/inventory/dcim/modules/',
            attrs={
                'data-static-params': '[{"queryParam":"has_asset_assigned","queryValue":"false"}]',
            },
        )
    )

    fieldsets = (
        ('Asset', ('module_type', 'name')),
        ('Module', ('site', 'device', 'module')),
        ('Tenancy', ('tenant', 'contact')),
    )

    class Meta:
        model = Asset
        fields = ('module_type', 'name', 'site', 'device', 'module', 'tenant', 'contact')
        widgets = {'module_type': forms.HiddenInput()}

    def clean_device(self):
        # prevents trying to set asset.device
        return None

    def clean_module_type(self):
        return self._clean_hardware_type('module')

    def clean_module(self):
        return self._clean_hardware('module')


class AssetInventoryItemAssignForm(AssetAssignMixin, NetBoxModelForm):
    site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        required=False,
        initial_params={'devices': '$device'},
    )
    device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        query_params={'site_id': '$site'},
        label='Device',
        required=False,
    )
    inventoryitem = DynamicModelChoiceField(
        queryset=InventoryItem.objects.all(),
        # we can't filter on inventoryitem_type because inventoryitem doesn't
        # have relation to inventoryitem_type
        query_params={'device_id': '$device'},
        label='Inventory item',
        required=False,
        help_text='Set to empty to unassign asset from inventory item',
        widget=APISelect(
            api_url='/api/plugins/inventory/dcim/inventory-items/',
            attrs={
                'data-static-params': '[{"queryParam":"has_asset_assigned","queryValue":"false"}]',
            },
        )
    )

    fieldsets = (
        ('Asset', ('inventoryitem_type', 'name')),
        ('Inventory Item', ('site', 'device', 'inventoryitem')),
        ('Tenancy', ('tenant', 'contact')),
    )

    class Meta:
        model = Asset
        fields = ('inventoryitem_type', 'name', 'site', 'device', 'inventoryitem', 'tenant', 'contact')
        widgets = {'inventoryitem_type': forms.HiddenInput()}

    def clean_device(self):
        # prevents trying to set asset.device
        return None

    def clean_inventoryitem_type(self):
        return self._clean_hardware_type('inventoryitem')

    def clean_inventoryitem(self):
        inventoryitem = self.cleaned_data['inventoryitem']
        if inventoryitem:
            if self.instance.inventoryitem == inventoryitem:
                # field was not changed
                return inventoryitem
            
            if Asset.objects.filter(inventoryitem=inventoryitem).exists():
                raise forms.ValidationError(f'Inventory item {inventoryitem} already has asset assigned')

        self.instance.inventoryitem = inventoryitem
        return inventoryitem



from copy import copy
from django.db.models import Count, F

from .choices import AssetStatusChoices
from .models import Asset


def asset_counts_type_status(inventoryitem_group, assets=None):
    """
    Return counts of assets based on combinations of inventoryitem type
    and status values for assets that belong to an inventoryitem group.
    Can optionally accept pre-filtered queryset with assets.
    Return value is a list of dicts, each having keys:
        - inventoryitem_type__manufacturer__name
        - inventoryitem_type__model
        - inventoryitem_type (ID)
        - status
        - label (display label of status)
        - color (color of status)
    """
    if assets is None:
        assets = Asset.objects.all()
    assets = assets.filter(
        inventoryitem_type__inventoryitem_group__in=inventoryitem_group.get_descendants(include_self=True)
    )
    # generate counts of assets grouped by type and status
    asset_counts = assets.values(
        'inventoryitem_type__manufacturer__name',
        'inventoryitem_type__model',
        'inventoryitem_type',
        'status',
    ).annotate(
        count=Count('pk')
    ).order_by('inventoryitem_type', 'status')

    def _update_status_meta(entry):
        """ adds color and label keys based on status value """
        entry['color'] = AssetStatusChoices.colors.get(entry['status'], 'gray')
        entry['label'] = dict(AssetStatusChoices).get(entry['status'], entry['status'])

    def _generate_entry(entry_from, status, count=0):
        t = copy(entry_from)
        t['status'] = status
        t['count'] = count
        _update_status_meta(t)
        return t

    # for each inventoryitem_type keep track of seen statues and add any that are
    # missing with count:0 
    zero_counts = []
    all_statuses = set(AssetStatusChoices.values())
    last_iid_pk = None
    seen_statues = set()
    seen_iit_pks = set()
    for idx, iit_status_count in enumerate(asset_counts):
        _update_status_meta(iit_status_count)
        seen_iit_pks.add(iit_status_count['inventoryitem_type'])
        if last_iid_pk is None:
            last_iid_pk = iit_status_count['inventoryitem_type']
        if last_iid_pk != iit_status_count['inventoryitem_type']:
            # next iit_pk, add unseen statuses of previous pk
            for missing_status in all_statuses - seen_statues:
                zero_counts.append(_generate_entry(asset_counts[idx-1], missing_status))
            # reset
            seen_statues = set()
        last_iid_pk = iit_status_count['inventoryitem_type']
        seen_statues.add(iit_status_count['status'])
    # complete missing statues for the last inventoryitem_type in asset_counts
    if last_iid_pk:
        for missing_status in all_statuses - seen_statues:
            zero_counts.append(_generate_entry(iit_status_count, missing_status))

    # now add entries for inventory item types that have no assets at all
    for iit in inventoryitem_group.inventoryitem_types.exclude(pk__in=seen_iit_pks).annotate(
        inventoryitem_type__manufacturer__name=F('manufacturer__name'),
        inventoryitem_type__model=F('model'),
        inventoryitem_type=F('pk'),
    ).values(
        'inventoryitem_type__manufacturer__name',
        'inventoryitem_type__model',
        'inventoryitem_type'
    ):
        for status in all_statuses:
            zero_counts.append(_generate_entry(iit, status))

    # combine non-zero and zero counts and sort
    asset_counts = sorted(
        list(asset_counts) + zero_counts,
        key=lambda k: (k['inventoryitem_type__manufacturer__name'], k['inventoryitem_type__model'], AssetStatusChoices.values().index(k['status']))
    )
    return asset_counts


def asset_counts_status(asset_counts):
    """
    Aggregates asset counts broken down by inventory item type and status
    (as returned by asset_counts_type_status) to counts on just status valuies.
    """
    status_counts = {
        k: {
            'value': k,
            'label': l,
            'color': AssetStatusChoices.colors[k],
            'count': sum(map(
                lambda e: e['count'],
                filter(lambda e: e['status']==k, asset_counts)
            ))
        }
        for k,l in list(AssetStatusChoices)
    }
    return status_counts



from dcim.api.views import DeviceViewSet, InventoryItemViewSet, ModuleViewSet
from netbox.api.viewsets import NetBoxModelViewSet
from utilities.utils import count_related
from .. import filtersets, models
from .serializers import (
    AssetSerializer, InventoryItemTypeSerializer, InventoryItemGroupSerializer,
    PurchaseSerializer, SupplierSerializer
)


class AssetViewSet(NetBoxModelViewSet):
    queryset = models.Asset.objects.prefetch_related(
        'device_type', 'device', 'module_type', 'module', 'storage_location', 'purchase__supplier', 'tags'
    )
    serializer_class = AssetSerializer
    filterset_class = filtersets.AssetFilterSet


class SupplierViewSet(NetBoxModelViewSet):
    queryset = models.Supplier.objects.prefetch_related('tags').annotate(
        asset_count=count_related(models.Asset, 'purchase__supplier'),
        purchase_count=count_related(models.Purchase, 'supplier'),
    )
    serializer_class = SupplierSerializer
    filterset_class = filtersets.SupplierFilterSet


class PurchaseViewSet(NetBoxModelViewSet):
    queryset = models.Purchase.objects.prefetch_related('tags').annotate(
        asset_count=count_related(models.Asset, 'purchase')
    )
    serializer_class = PurchaseSerializer
    filterset_class = filtersets.PurchaseFilterSet


class InventoryItemTypeViewSet(NetBoxModelViewSet):
    queryset = models.InventoryItemType.objects.prefetch_related('tags').annotate(
        asset_count=count_related(models.Asset, 'inventoryitem_type')
    )
    serializer_class = InventoryItemTypeSerializer
    filterset_class = filtersets.InventoryItemTypeFilterSet


class InventoryItemGroupViewSet(NetBoxModelViewSet):
    queryset = models.InventoryItemGroup.objects.add_related_count(
        models.InventoryItemGroup.objects.all(),
        models.Asset,
        'inventoryitem_type__inventoryitem_group',
        'asset_count',
        cumulative=True
    ).prefetch_related('tags')
    serializer_class = InventoryItemGroupSerializer
    filterset_class = filtersets.InventoryItemGroupFilterSet


class DeviceAssetViewSet(DeviceViewSet):
    """Adds option to filter on asset assignemnet"""
    filterset_class = filtersets.DeviceAssetFilterSet


class ModuleAssetViewSet(ModuleViewSet):
    """Adds option to filter on asset assignemnet"""
    filterset_class = filtersets.ModuleAssetFilterSet


class InventoryItemAssetViewSet(InventoryItemViewSet):
    """Adds option to filter on asset assignemnet"""
    filterset_class = filtersets.InventoryItemAssetFilterSet



from utilities.testing import APIViewTestCases
from ..custom import APITestCase
from ...models import InventoryItemGroup


class InventoryItemGroupTest(
        APITestCase, 
        APIViewTestCases.GetObjectViewTestCase,
        APIViewTestCases.ListObjectsViewTestCase,
        APIViewTestCases.CreateObjectViewTestCase,
        APIViewTestCases.UpdateObjectViewTestCase,
        APIViewTestCases.DeleteObjectViewTestCase):
    model = InventoryItemGroup
    brief_fields = ['_depth', 'display', 'id', 'name', 'url']
    create_data = [
        {
            'name': 'InventoryItemGroup 4',
        },
        {
            'name': 'InventoryItemGroup 5',
        },
        {
            'name': 'InventoryItemGroup 6',
        },
    ]

    @classmethod
    def setUpTestData(cls) -> None:
        InventoryItemGroup.objects.create(name='InventoryItemGroup 1')
        InventoryItemGroup.objects.create(name='InventoryItemGroup 2')
        InventoryItemGroup.objects.create(name='InventoryItemGroup 3')



from dcim.models import Manufacturer
from utilities.testing import ViewTestCases

from netbox_inventory.tests.custom import ModelViewTestCase
from netbox_inventory.models import InventoryItemType, InventoryItemGroup


class InventoryItemTypeTestCase(
    ModelViewTestCase,
    ViewTestCases.PrimaryObjectViewTestCase,
):

    model = InventoryItemType

    @classmethod
    def setUpTestData(cls):
        manufacturer1 = Manufacturer.objects.create(
            name='Manufacturer 1',
            slug='manufacturer1',
        )
        manufacturer2 = Manufacturer.objects.create(
            name='Manufacturer 2',
            slug='manufacturer2',
        )
        inventoryitem_group1 = InventoryItemGroup.objects.create(name='IIG1')
        inventoryitemtype1 = InventoryItemType.objects.create(
            model='InventoryItemType 1',
            slug='inventoryitemtype1',
            manufacturer=manufacturer1,
        )
        inventoryitemtype2 = InventoryItemType.objects.create(
            model='InventoryItemType 2',
            slug='inventoryitemtype2',
            manufacturer=manufacturer1,
        )
        inventoryitemtype3 = InventoryItemType.objects.create(
            model='InventoryItemType 3',
            slug='inventoryitemtype3',
            manufacturer=manufacturer1,
        )
        cls.form_data = {
            'model': 'InventoryItemType',
            'slug': 'inventoryitemtype',
            'manufacturer': manufacturer1.pk,
            'part_number': 'InventoryItemType PN',
        }
        cls.csv_data = (
            'model,slug,manufacturer',
            f'InventoryItemType 4,inventoryitemtype4,{manufacturer1.name}',
            f'InventoryItemType 5,inventoryitemtype5,{manufacturer1.name}',
            f'InventoryItemType 6,inventoryitemtype6,{manufacturer1.name}',
        )
        cls.csv_update_data = (
            'id,manufacturer',
            f'{inventoryitemtype1.pk},{manufacturer2.name}',
            f'{inventoryitemtype2.pk},{manufacturer2.name}',
            f'{inventoryitemtype3.pk},{manufacturer2.name}',
        )
        cls.bulk_edit_data = {
            'inventoryitem_group': inventoryitem_group1.pk,
        }



from dcim.models import Device, InventoryItem, Module
from netbox.views import generic
from ..forms.create import *
from ..models import Asset

__all__ = (
    'AssetDeviceCreateView',
    'AssetModuleCreateView',
    'AssetInventoryItemCreateView',
)


class AssetCreateView(generic.ObjectEditView):
    template_name = 'netbox_inventory/asset_create.html'
    asset = None

    def _load_asset(self, request):
        asset_id = request.GET.get('asset_id')
        if asset_id:
            self.asset = Asset.objects.get(pk=asset_id)

    def dispatch(self, request, *args, **kwargs):
        self._load_asset(request)
        return super().dispatch(request, *args, **kwargs)

    def alter_object(self, obj, request, url_args, url_kwargs):
        obj.assigned_asset = self.asset
        return super().alter_object(obj, request, url_args, url_kwargs)

    def get_extra_context(self, request, instance):
        context = super().get_extra_context(request, instance)
        context['asset'] = self.asset
        return context


class AssetDeviceCreateView(AssetCreateView):
    queryset = Device.objects.all()
    form = AssetDeviceCreateForm

    def get_object(self, **kwargs):
        return Device(assigned_asset=self.asset)

    def get_extra_context(self, request, instance):
        context = super().get_extra_context(request, instance)
        context['template_extends'] = 'dcim/device_edit.html'
        return context


class AssetModuleCreateView(AssetCreateView):
    queryset = Module.objects.all()
    form = AssetModuleCreateForm

    def get_object(self, **kwargs):
        return Module(assigned_asset=self.asset)


class AssetInventoryItemCreateView(AssetCreateView):
    queryset = InventoryItem.objects.all()
    form = AssetInventoryItemCreateForm

    def get_object(self, **kwargs):
        return InventoryItem(assigned_asset=self.asset)



from netbox.views import generic

from ..forms.assign import *
from ..models import Asset

__all__ = (
    'AssetAssignView',
)


class AssetAssignView(generic.ObjectEditView):
    queryset = Asset.objects.all()
    template_name = 'netbox_inventory/asset_assign.html'

    def dispatch(self, request, *args, **kwargs):
        # Set the form class based on the type of hardware being assigned
        obj = self.get_object(**kwargs)
        self.form = {
            'device': AssetDeviceAssignForm,
            'module': AssetModuleAssignForm,
            'inventoryitem': AssetInventoryItemAssignForm,
        }[obj.kind]
        return super().dispatch(request, *args, **kwargs)



from netbox.views import generic
from utilities.utils import count_related
from .. import filtersets, forms, models, tables

__all__ = (
    'InventoryItemTypeView',
    'InventoryItemTypeListView',
    'InventoryItemTypeEditView',
    'InventoryItemTypeDeleteView',
    'InventoryItemTypeBulkImportView',
    'InventoryItemTypeBulkEditView',
    'InventoryItemTypeBulkDeleteView',
)

class InventoryItemTypeView(generic.ObjectView):
    queryset = models.InventoryItemType.objects.all()


class InventoryItemTypeListView(generic.ObjectListView):
    queryset = models.InventoryItemType.objects.annotate(
        asset_count=count_related(models.Asset, 'inventoryitem_type'),
    )
    table = tables.InventoryItemTypeTable
    filterset = filtersets.InventoryItemTypeFilterSet
    filterset_form = forms.InventoryItemTypeFilterForm


class InventoryItemTypeEditView(generic.ObjectEditView):
    queryset = models.InventoryItemType.objects.all()
    form = forms.InventoryItemTypeForm


class InventoryItemTypeDeleteView(generic.ObjectDeleteView):
    queryset = models.InventoryItemType.objects.all()


class InventoryItemTypeBulkImportView(generic.BulkImportView):
    queryset = models.InventoryItemType.objects.all()   
    table = tables.InventoryItemTypeTable 
    model_form = forms.InventoryItemTypeImportForm


class InventoryItemTypeBulkEditView(generic.BulkEditView):
    queryset = models.InventoryItemType.objects.all()
    filterset = filtersets.InventoryItemTypeFilterSet
    table = tables.InventoryItemTypeTable
    form = forms.InventoryItemTypeBulkEditForm


class InventoryItemTypeBulkDeleteView(generic.BulkDeleteView):
    queryset = models.InventoryItemType.objects.all()
    table = tables.InventoryItemTypeTable



from dcim.models import Manufacturer
from utilities.testing import APIViewTestCases
from ..custom import APITestCase
from ...models import InventoryItemType, InventoryItemGroup


class InventoryItemTypeTest(
        APITestCase, 
        APIViewTestCases.GetObjectViewTestCase,
        APIViewTestCases.ListObjectsViewTestCase,
        APIViewTestCases.CreateObjectViewTestCase,
        APIViewTestCases.UpdateObjectViewTestCase,
        APIViewTestCases.DeleteObjectViewTestCase):
    model = InventoryItemType
    brief_fields = ['display', 'id', 'manufacturer', 'model', 'slug', 'url']


    @classmethod
    def setUpTestData(cls) -> None:
        manufacturer1 = Manufacturer.objects.create(name='Manufacturer 1', slug='manufacturer1')
        ig1 = InventoryItemGroup.objects.create(name='IG1')
        InventoryItemType.objects.create(model='InventoryItemType 1', slug='inventoryitemtype1', manufacturer=manufacturer1)
        InventoryItemType.objects.create(model='InventoryItemType 2', slug='inventoryitemtype2', manufacturer=manufacturer1)
        InventoryItemType.objects.create(model='InventoryItemType 3', slug='inventoryitemtype3', manufacturer=manufacturer1)
        cls.create_data = [
            {
                'model': 'InventoryItemType 4',
                'slug': 'inventoryitemtype4',
                'manufacturer': manufacturer1.pk,
            },
            {
                'model': 'InventoryItemType 5',
                'slug': 'inventoryitemtype5',
                'manufacturer': manufacturer1.pk,
            },
            {
                'model': 'InventoryItemType 6',
                'slug': 'inventoryitemtype6',
                'manufacturer': manufacturer1.pk,
            },
        ]
        cls.bulk_update_data = {
            'inventoryitem_group': ig1.pk,
        }



# Generated by Django 4.0.8 on 2022-12-11 12:44

from django.db import migrations, models
import django.db.models.deletion
import taggit.managers
from utilities.json import CustomFieldJSONEncoder


class Migration(migrations.Migration):

    dependencies = [
        ('extras', '0077_customlink_extend_text_and_url'),
        ('netbox_inventory', '0002_alter_asset_serial'),
    ]

    operations = [
        migrations.CreateModel(
            name='InventoryItemGroup',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=CustomFieldJSONEncoder)),
                ('name', models.CharField(max_length=100, unique=True)),
                ('comments', models.TextField(blank=True)),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.AddField(
            model_name='inventoryitemtype',
            name='inventoryitem_group',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='inventoryitem_types', to='netbox_inventory.inventoryitemgroup'),
        ),
    ]



from extras.plugins import PluginConfig
from .version import __version__


class NetBoxInventoryConfig(PluginConfig):
    name = 'netbox_inventory'
    verbose_name = 'NetBox Inventory'
    version = __version__
    description = 'Inventory asset management in NetBox'
    author = 'Matej Vadnjal'
    author_email = 'matej.vadnjal@arnes.si'
    base_url = 'inventory'
    min_version = '3.3.0'
    default_settings = {
        'top_level_menu': True,
        'used_status_name': 'used',
        'stored_status_name': 'stored',
        'sync_hardware_serial_asset_tag': False,
        'asset_import_create_purchase': False,
        'asset_import_create_device_type': False,
        'asset_import_create_module_type': False,
        'asset_import_create_inventoryitem_type': False,
        'asset_disable_editing_fields_for_tags': {},
        'asset_disable_deletion_for_tags': [],
    }

    def ready(self):
        super().ready()
        import netbox_inventory.signals

config = NetBoxInventoryConfig



from utilities.testing import ViewTestCases

from netbox_inventory.tests.custom import ModelViewTestCase
from netbox_inventory.models import InventoryItemGroup


class InventoryItemGroupTestCase(
    ModelViewTestCase,
    ViewTestCases.PrimaryObjectViewTestCase,
):

    model = InventoryItemGroup

    form_data = {
        'name': 'InventoryItemGroup',
    }
    csv_data = (
        'name,comments',
        'IIG4,a comment',
        'IIG5,a comment',
        'IIG6,a comment',
    )

    @classmethod
    def setUpTestData(cls):
        iig1 = InventoryItemGroup.objects.create(name='IIG1')
        iig2 = InventoryItemGroup.objects.create(name='IIG2')
        iig3 = InventoryItemGroup.objects.create(name='IIG3')

        cls.csv_update_data = (
            'id,name,parent',
            f'{iig1.pk},IIG1_update,InventoryItemGroup',
            f'{iig2.pk},IIG2_update,InventoryItemGroup',
            f'{iig3.pk},IIG3_update,InventoryItemGroup',
        )
        cls.bulk_edit_data = {
            'comments': 'updated',
        }


from utilities.testing import APIViewTestCases
from ..custom import APITestCase
from ...models import Purchase, Supplier


class PurchaseTest(
        APITestCase, 
        APIViewTestCases.GetObjectViewTestCase,
        APIViewTestCases.ListObjectsViewTestCase,
        APIViewTestCases.CreateObjectViewTestCase,
        APIViewTestCases.UpdateObjectViewTestCase,
        APIViewTestCases.DeleteObjectViewTestCase):
    model = Purchase
    brief_fields = ['date', 'display', 'id', 'name', 'supplier', 'url']

    bulk_update_data = {
        'description': 'new description',
    }

    @classmethod
    def setUpTestData(cls) -> None:
        supplier1 = Supplier.objects.create(name='Supplier 1')
        Purchase.objects.create(name='Purchase 1', supplier=supplier1)
        Purchase.objects.create(name='Purchase 2', supplier=supplier1)
        Purchase.objects.create(name='Purchase 3', supplier=supplier1)
        cls.create_data = [
            {
                'name': 'Purchase 4',
                'supplier': supplier1.pk,
            },
            {
                'name': 'Purchase 5',
                'supplier': supplier1.pk,
            },
            {
                'name': 'Purchase 6',
                'supplier': supplier1.pk,
            },
        ]



from utilities.choices import ChoiceSet

class AssetStatusChoices(ChoiceSet):
    key = 'Asset.status'

    CHOICES = [
        ('stored', 'Stored', 'green'),
        ('used', 'Used', 'blue'),
        ('retired', 'Retired', 'gray'),
    ]


class HardwareKindChoices(ChoiceSet):
    CHOICES = [
        ('device', 'Device'),
        ('module', 'Module'),
        ('inventoryitem', 'Inventory Item'),
    ]



from utilities.testing import ViewTestCases

from netbox_inventory.tests.custom import ModelViewTestCase
from netbox_inventory.models import Supplier


class SupplierTestCase(
    ModelViewTestCase,
    ViewTestCases.PrimaryObjectViewTestCase,
):

    model = Supplier

    form_data = {
        'name': 'Supplier',
        'slug': 'supplier',
        'description': 'supplier description',
    }
    csv_data = (
        'name,slug',
        'Supplier 4,supplier4',
        'Supplier 5,supplier5',
        'Supplier 6,supplier6',
    )
    bulk_edit_data = {
        'description': 'bulk description',
    }

    @classmethod
    def setUpTestData(cls):
        supplier1 = Supplier.objects.create(
            name='Supplier 1',
            slug='supplier1',
        )
        supplier2 = Supplier.objects.create(
            name='Supplier 2',
            slug='supplier2',
        )
        supplier3 = Supplier.objects.create(
            name='Supplier 3',
            slug='supplier3',
        )
        cls.csv_update_data = (
            'id,description',
            f'{supplier1.pk},description 1',
            f'{supplier2.pk},description 2',
            f'{supplier3.pk},description 3',
        )



import logging

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError

from dcim.forms import DeviceForm, InventoryItemForm, ModuleForm
from dcim.models.device_components import ConsolePort, ConsoleServerPort, FrontPort, Interface, PowerOutlet, PowerPort, RearPort
from utilities.forms import DynamicModelChoiceField, StaticSelect

__all__ = (
    'AssetDeviceCreateForm',
    'AssetModuleCreateForm',
    'AssetInventoryItemCreateForm',
)


COMPONENT_FIELDS = (
    'consoleport',
    'consoleserverport',
    'frontport',
    'interface',
    'poweroutlet',
    'powerport',
    'rearport',
)

logger = logging.getLogger('netbox.netbox_inventory.forms.create')


class AssetCreateMixin:
    def update_hardware_fields(self, kind_type):
        """
            Pre-populate and disable hardware related fields from asset data
        """
        if self.instance.assigned_asset:
            asset = self.instance.assigned_asset
            self.fields['serial'].disabled = True
            self.fields['asset_tag'].disabled = True
            self.initial['serial'] = asset.serial
            self.initial['asset_tag'] = asset.asset_tag if asset.asset_tag else None

            self.fields['manufacturer'].widget = StaticSelect(attrs={'readonly':True, 'disabled':True})
            self.fields['manufacturer'].choices = [(asset.hardware_type.manufacturer.pk, asset.hardware_type.manufacturer)]
            self.fields[kind_type].widget = StaticSelect(attrs={'readonly':True, 'disabled':True})
            self.fields[kind_type].choices = [(asset.hardware_type.pk, asset.hardware_type)]
            # disabled Select field will not be POSTed so clean() complains if field is required
            # we set value later in clean_*_type() anyway
            self.fields[kind_type].required = False

    def save(self, *args):
        """
        After we save new hardware (Device, Module, InventortyItem), we must update
        asset.device/.module/.intentory_item to reffer to this new hardware instance
        """
        asset = self.instance.assigned_asset
        instance = super().save(*args)
        asset.snapshot()
        setattr(asset, asset.kind, instance)
        asset.full_clean()
        asset.save()
        logger.info(f'Assigned newly created hardware ({instance}) to asset {asset}')
        return instance


class AssetDeviceCreateForm(AssetCreateMixin, DeviceForm):
    """
        Populates and disables editing of asset and devcie_type fields
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.update_hardware_fields('device_type')

    def clean_device_type(self):
        # no mattter what was POSTed, device_type cannot be changed/missing...
        return self.instance.assigned_asset.device_type


class AssetModuleCreateForm(AssetCreateMixin, ModuleForm):
    """
        Populates and disables editing of asset and module_type fields
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.update_hardware_fields('module_type')

    def clean_module_type(self):
        return self.instance.assigned_asset.module_type


class AssetInventoryItemCreateForm(AssetCreateMixin, InventoryItemForm):
    """
        Populates and disables editing of hardware related fields
        Offers selection of device components and maps selected component
        to component_type and component_id fields
    """

    consoleport = DynamicModelChoiceField(
        queryset=ConsolePort.objects.all(),
        query_params={'device_id': '$device'},
        label='Console port',
        required=False,
    )
    consoleserverport = DynamicModelChoiceField(
        queryset=ConsoleServerPort.objects.all(),
        query_params={'device_id': '$device'},
        label='Console server port',
        required=False,
    )
    frontport = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        query_params={'device_id': '$device'},
        label='Front port',
        required=False,
    )
    interface = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        query_params={'device_id': '$device'},
        label='Interface',
        required=False,
    )
    poweroutlet = DynamicModelChoiceField(
        queryset=PowerOutlet.objects.all(),
        query_params={'device_id': '$device'},
        label='Power outlet',
        required=False,
    )
    powerport = DynamicModelChoiceField(
        queryset=PowerPort.objects.all(),
        query_params={'device_id': '$device'},
        label='Power port',
        required=False,
    )
    rearport = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        query_params={'device_id': '$device'},
        label='Rear port',
        required=False,
    )

    fieldsets = (
        InventoryItemForm.fieldsets[0],
        ('Component', ('interface', 'consoleport', 'consoleserverport', 'frontport', 'rearport', 'poweroutlet', 'powerport')),
    ) + InventoryItemForm.fieldsets[1:]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.instance.assigned_asset:
            asset = self.instance.assigned_asset
            self.fields['serial'].disabled = True
            self.fields['asset_tag'].disabled = True
            self.fields['part_id'].disabled = True
            self.initial['serial'] = asset.serial
            self.initial['asset_tag'] = asset.asset_tag if asset.asset_tag else None
            self.initial['part_id'] = asset.inventoryitem_type.part_number

            self.fields['manufacturer'].widget = StaticSelect(attrs={'readonly':True, 'disabled':True})
            self.fields['manufacturer'].choices = [(asset.inventoryitem_type.manufacturer.pk, asset.inventoryitem_type.manufacturer)]

    def clean(self):
        super().clean()
        component_set = None
        for field_name in COMPONENT_FIELDS:
            field_value = self.cleaned_data.get(field_name)
            if field_value and component_set:
                raise ValidationError('Only a single component can be selected')
            if field_value:
                component_set = field_name
                self.cleaned_data['component_type'] = ContentType.objects.get(app_label='dcim', model=field_name)
                self.cleaned_data['component_id'] = field_value.pk
                self.cleaned_data.pop(field_name)

    def clean_manufacturer(self):
        return self.instance.assigned_asset.inventoryitem_type.manufacturer



from netbox.search import SearchIndex
from .models import Asset, InventoryItemType, InventoryItemGroup, Purchase, Supplier


class AssetIndex(SearchIndex):
    model = Asset
    fields = (
        ("name", 100),
        ("asset_tag", 50),
        ("serial", 60),
        ("comments", 5000),
    )


class SupplierIndex(SearchIndex):
    model = Supplier
    fields = (
        ("name", 100),
        ("description", 500),
        ("comments", 5000),
    )


class PurchaseIndex(SearchIndex):
    model = Purchase
    fields = (
        ("name", 100),
        ("description", 500),
        ("comments", 5000),
    )


class InventoryItemTypeIndex(SearchIndex):
    model = InventoryItemType
    fields = (
        ("model", 100),
        ("part_number", 100),
        ("comments", 5000),
    )


class InventoryItemGroupIndex(SearchIndex):
    model = InventoryItemGroup
    fields = (
        ("name", 100),
        ("comments", 5000),
    )


indexes = [
    AssetIndex,
    SupplierIndex,
    PurchaseIndex,
    InventoryItemTypeIndex,
    InventoryItemGroupIndex,
]



from django.urls import reverse
from django.test import SimpleTestCase

from netbox_inventory import __version__
from netbox_inventory.tests.custom import APITestCase


class NetboxDnsVersionTestCase(SimpleTestCase):
    """
    Test for netbox_inventory package
    """

    def test_version(self):
        assert __version__ == "1.2.1"


class AppTest(APITestCase):
    """
    Test the availability of the plugin API root
    """

    def test_root(self):
        url = reverse("plugins-api:netbox_inventory-api:api-root")
        response = self.client.get(f"{url}?format=api", **self.header)

        self.assertEqual(response.status_code, 200)



from django import forms

from utilities.forms import BootstrapMixin

__all__ = (
    'AssetBulkAddForm',
)


class AssetBulkAddForm(BootstrapMixin, forms.Form):
    """ Form for creating multiple Assets by count """
    count = forms.IntegerField(
        min_value=1,
        required=True,
        help_text='How many assets to create',
    )



from netbox.views import generic

from .. import filtersets, forms, models, tables
from ..choices import AssetStatusChoices
from ..analyzers import asset_counts_type_status, asset_counts_status

__all__ = (
    'InventoryItemGroupView',
    'InventoryItemGroupListView',
    'InventoryItemGroupEditView',
    'InventoryItemGroupDeleteView',
    'InventoryItemGroupBulkImportView',
    'InventoryItemGroupBulkEditView',
    'InventoryItemGroupBulkDeleteView',
)


class InventoryItemGroupView(generic.ObjectView):
    queryset = models.InventoryItemGroup.objects.all()

    def get_extra_context(self, request, instance):
        # build a table fo child groups with asset count
        child_groups = models.InventoryItemGroup.objects.add_related_count(
            models.InventoryItemGroup.objects.all(),
            models.Asset,
            'inventoryitem_type__inventoryitem_group',
            'asset_count',
            cumulative=True
        )
        child_groups = models.InventoryItemGroup.objects.add_related_count(
            child_groups,
            models.InventoryItemType,
            'inventoryitem_group',
            'inventoryitem_type_count',
            cumulative=True
        ).restrict(request.user, 'view').filter(
            parent__in=instance.get_descendants(include_self=True)
        )
        child_groups_table = tables.InventoryItemGroupTable(child_groups)
        child_groups_table.columns.hide('actions')
        # get all assets from this group and its descendants
        assets = models.Asset.objects.restrict(request.user, 'view').filter(
            inventoryitem_type__inventoryitem_group__in=instance.get_descendants(include_self=True)
        )
        # make table of assets
        asset_table = tables.AssetTable(assets, user=request.user)
        asset_table.columns.hide('kind')
        asset_table.configure(request)

        # get counts for each inventoryitem type and status combination
        type_status_counts = asset_counts_type_status(instance, assets)
        
        # counts by status, ignoring different inventoryitem_types
        status_counts = asset_counts_status(type_status_counts)

        return {
            'child_groups_table': child_groups_table,
            'asset_table': asset_table,
            'type_status_counts': type_status_counts,
            'status_counts': status_counts,
        }


class InventoryItemGroupListView(generic.ObjectListView):
    queryset = models.InventoryItemGroup.objects.add_related_count(
        models.InventoryItemGroup.objects.all(),
        models.Asset,
        'inventoryitem_type__inventoryitem_group',
        'asset_count',
        cumulative=True
    )
    queryset = models.InventoryItemGroup.objects.add_related_count(
        queryset,
        models.InventoryItemType,
        'inventoryitem_group',
        'inventoryitem_type_count',
        cumulative=True
    )
    table = tables.InventoryItemGroupTable
    filterset = filtersets.InventoryItemGroupFilterSet
    filterset_form = forms.InventoryItemGroupFilterForm


class InventoryItemGroupEditView(generic.ObjectEditView):
    queryset = models.InventoryItemGroup.objects.all()
    form = forms.InventoryItemGroupForm


class InventoryItemGroupDeleteView(generic.ObjectDeleteView):
    queryset = models.InventoryItemGroup.objects.all()


class InventoryItemGroupBulkImportView(generic.BulkImportView):
    queryset = models.InventoryItemGroup.objects.all()   
    table = tables.InventoryItemGroupTable 
    model_form = forms.InventoryItemGroupImportForm


class InventoryItemGroupBulkEditView(generic.BulkEditView):
    queryset = models.InventoryItemGroup.objects.all()
    filterset = filtersets.InventoryItemGroupFilterSet
    table = tables.InventoryItemGroupTable
    form = forms.InventoryItemGroupBulkEditForm


class InventoryItemGroupBulkDeleteView(generic.BulkDeleteView):
    queryset = models.InventoryItemGroup.objects.all()
    table = tables.InventoryItemGroupTable



from datetime import date

from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.forms import ValidationError
from django.urls import reverse

from netbox.models import NetBoxModel, NestedGroupModel
from .choices import HardwareKindChoices, AssetStatusChoices
from .utils import asset_clear_old_hw, get_prechange_field, get_plugin_setting, get_status_for


class Asset(NetBoxModel):
    """
    An Asset represents a piece of hardware we want to keep track of. It has a
    make (model, part number) that is one of: Device Type, Module Type or
    InventoryItem Type.

    Asset must have a serial number, can have an asset tag (inventory number). It
    must have one of DeviceType, ModuleType or InventoryItemType. It can have a
    storage location (instance of Location). There are also fields to keep track of
    purchase and warranty info.

    An asset that is in use, can be assigned to a Device, Module or InventoryItem.
    """
    #
    # fields that identify asset
    #
    name = models.CharField(
        help_text='Can be used to quickly identify a particular asset',
        max_length=128,
        blank=True,
        null=False,
        default='',
    )
    asset_tag = models.CharField(
        help_text='Identifier assigned by owner',
        max_length=50,
        blank=True,
        null=True,
        default=None,
        verbose_name='Asset Tag',
    )
    serial = models.CharField(
        help_text='Identifier assigned by manufacturer',
        max_length=60,
        verbose_name='Serial Number',
        blank=True,
        null=True,
        default=None,
    )

    #
    # status fields
    #
    status = models.CharField(
        max_length=30,
        choices=AssetStatusChoices,
        help_text='Asset lifecycle status',
    )

    #
    # hardware type fields
    #
    device_type = models.ForeignKey(
        to='dcim.DeviceType',
        on_delete=models.PROTECT,
        related_name='+',
        blank=True,
        null=True,
        verbose_name='Device Type',
    )
    module_type = models.ForeignKey(
        to='dcim.ModuleType',
        on_delete=models.PROTECT,
        related_name='+',
        blank=True,
        null=True,
        verbose_name='Module Type',
    )
    inventoryitem_type = models.ForeignKey(
        to='netbox_inventory.InventoryItemType',
        on_delete=models.PROTECT,
        related_name='+',
        blank=True,
        null=True,
        verbose_name='Inventory Item Type',
    )

    #
    # used fields
    #
    device = models.OneToOneField(
        to='dcim.Device',
        on_delete=models.SET_NULL,
        related_name='assigned_asset',
        blank=True,
        null=True,
    )
    module = models.OneToOneField(
        to='dcim.Module',
        on_delete=models.SET_NULL,
        related_name='assigned_asset',
        blank=True,
        null=True,
    )
    inventoryitem = models.OneToOneField(
        to='dcim.InventoryItem',
        on_delete=models.SET_NULL,
        related_name='assigned_asset',
        blank=True,
        null=True,
        verbose_name='Inventory Item',
    )
    tenant = models.ForeignKey(
        help_text='Tenant using this asset',
        to='tenancy.Tenant',
        on_delete=models.PROTECT,
        related_name='+',
        blank=True,
        null=True,
    )
    contact = models.ForeignKey(
        help_text='Contact using this asset',
        to='tenancy.Contact',
        on_delete=models.PROTECT,
        related_name='+',
        blank=True,
        null=True,
    )

    storage_location = models.ForeignKey(
        help_text='Where is this asset stored when not in use',
        to='dcim.Location',
        on_delete=models.PROTECT,
        related_name='+',
        blank=True,
        null=True,
        verbose_name='Storage Location',
    )

    #
    # purchase info
    #
    owner = models.ForeignKey(
        help_text='Who owns this asset',
        to='tenancy.Tenant',
        on_delete=models.PROTECT,
        related_name='+',
        blank=True,
        null=True,
    )
    purchase = models.ForeignKey(
        help_text='Purchase through which this asset was purchased',
        to='netbox_inventory.Purchase',
        on_delete=models.PROTECT,
        related_name='assets',
        blank=True,
        null=True,
    )
    warranty_start = models.DateField(
        help_text='First date warranty for this asset is valid',
        blank=True,
        null=True,
        verbose_name='Warranty Start',
    )
    warranty_end = models.DateField(
        help_text='Last date warranty for this asset is valid',
        blank=True,
        null=True,
        verbose_name='Warranty End',
    )

    comments = models.TextField(
        blank=True
    )

    images = GenericRelation(
        to='extras.ImageAttachment'
    )

    clone_fields = [
        'status', 'device_type', 'module_type', 'inventoryitem_type', 'tenant', 'contact',
        'storage_location', 'owner', 'purchase', 'warranty_start', 'warranty_end', 'comments'
    ]

    @property
    def kind(self):
        if self.device_type_id:
            return 'device'
        elif self.module_type_id:
            return 'module'
        elif self.inventoryitem_type_id:
            return 'inventoryitem'
        assert False, f'Invalid hardware kind detected for asset {self.pk}'

    def get_kind_display(self):
        return dict(HardwareKindChoices)[self.kind]

    @property
    def hardware_type(self):
        return self.device_type or self.module_type or self.inventoryitem_type or None

    @property
    def hardware(self):
        return self.device or self.module or self.inventoryitem or None

    @property
    def installed_site(self):
        device = self.installed_device
        if device:
            return device.site

    @property
    def installed_location(self):
        device = self.installed_device
        if device:
            return device.location

    @property
    def installed_rack(self):
        device = self.installed_device
        if device:
            return device.rack

    @property
    def installed_device(self):
        if self.kind == 'device':
            return self.device
        elif self.hardware:
            return self.hardware.device
        else:
            return None

    @property
    def warranty_remaining(self):
        """
            How many days are left in warranty period.
            Returns negative duration if period has not started yet
            Return None if warranty_end not defined
        """
        if self.warranty_end:
            if self.warranty_start and self.warranty_start > date.today():
                return date.today() - self.warranty_start
            else:
                return self.warranty_end - date.today()
        return None

    @property
    def warranty_elapsed(self):
        """
            How many days have passed in warranty period.
            Returns negative duration if period has not started yet
            Return None if warranty_start not defined
        """
        if self.warranty_start:
            return date.today() - self.warranty_start
        return None

    @property
    def warranty_total(self):
        if self.warranty_end and self.warranty_start:
            return self.warranty_end - self.warranty_start
        return None

    def clean(self):
        self.validate_hardware_types()
        self.validate_hardware()
        self.update_status()
        return super().clean()

    def save(self, *args, **kwargs):
        self.update_hardware_used()
        return super().save(*args, **kwargs)

    def validate_hardware_types(self):
        """Ensure only one device/module_type/inventoryitem_type is set at a time."""
        if sum(map(bool, [self.device_type, self.module_type, self.inventoryitem_type])) > 1:
            raise ValidationError('Only one of device type, module type and inventory item type can be set for the same asset.')
        if not self.device_type and not self.module_type and not self.inventoryitem_type:
            raise ValidationError('One of device type, module type or inventory item type must be set.')

    def validate_hardware(self):
        """Ensure only one device/module is set at a time and it matches device/module_type."""
        kind = self.kind
        _type = getattr(self, kind+'_type')
        hw = getattr(self, kind)
        hw_others = dict(HardwareKindChoices).keys() - [kind]

        # e.g.: self.device_type and self.device.device_type must match
        # InventoryItem does not have FK to InventoryItemType
        if kind != 'inventoryitem' and hw and _type != getattr(hw, kind+'_type'):
            raise ValidationError({kind: f'{kind} type of {kind} does not match {kind} type of asset'})
        # ensure only one hardware is set and that it is correct kind
        # e.g. if self.device_type is set, we cannot have self.module or self.inventoryitem set
        for hw_other in hw_others:
            if getattr(self, hw_other):
                raise ValidationError(f'Cannot set {hw_other} for asset that is a {kind}')

    def update_status(self):
        """ If asset was assigned or unassigned to a particular device, module, inventoryitem
            update asset.status. Depending on plugin configuration.
        """
        old_hw = get_prechange_field(self, self.kind)
        new_hw = getattr(self, self.kind)
        old_status = get_prechange_field(self, 'status')
        stored_status = get_status_for('stored')
        used_status = get_status_for('used')
        if old_status != self.status:
            # status has also been changed manually, don't change it automatically
            return
        if used_status and new_hw and not old_hw:
            self.status = used_status
        elif stored_status and not new_hw and old_hw:
            self.status = stored_status

    def update_hardware_used(self):
        """ If assigning as device, module or inventoryitem set serial and
            asset_tag on it. Also remove them if unasigning.
        """
        if not get_plugin_setting('sync_hardware_serial_asset_tag'):
            return None
        old_hw = get_prechange_field(self, self.kind)
        new_hw = getattr(self, self.kind)
        old_serial = get_prechange_field(self, 'serial')
        old_asset_tag = get_prechange_field(self, 'asset_tag')
        # device, module... needs None for unset asset_tag to enforce uniqness at DB level
        new_asset_tag = self.asset_tag if self.asset_tag else None
        # device, module... does not allow serial to be null
        new_serial = self.serial if self.serial else ''
        if not new_hw and old_hw:
            # unassigned
            asset_clear_old_hw(old_hw)
        elif new_hw and old_hw != new_hw:
            # assigned something, set its serial
            if old_hw:
                # but first clear previous hw data
                asset_clear_old_hw(old_hw)
            # if new_hw already has correct values, don't save it again
            new_hw_save = False
            if new_hw.serial != new_serial:
                new_hw.serial = new_serial
                new_hw_save = True
            if new_hw.asset_tag != new_asset_tag:
                new_hw.asset_tag = new_asset_tag
                new_hw_save = True
            if new_hw_save:
                new_hw.save()
        elif self.serial != old_serial or self.asset_tag != old_asset_tag:
            # just changed asset's serial or asset_tag, update assigned hw
            if new_hw:
                new_hw.serial = new_serial
                new_hw.asset_tag = new_asset_tag
                new_hw.save()

    def get_absolute_url(self):
        return reverse('plugins:netbox_inventory:asset', args=[self.pk])

    def get_status_color(self):
        return AssetStatusChoices.colors.get(self.status)

    def __str__(self):
        return f'{self.hardware_type} {self.serial}'

    class Meta:
        ordering = ('device_type', 'module_type', 'inventoryitem_type', 'serial',)
        unique_together = (
            ('device_type', 'serial'),
            ('module_type', 'serial'),
            ('inventoryitem_type', 'serial'),
            ('owner', 'asset_tag'),
        )


class Supplier(NetBoxModel):
    """
    Supplier is a legal entity that sold some assets that we keep track of.
    This can be the same entity as Manufacturer or a separate one. However
    netbox_inventory keeps track of Suppliers separate from Manufacturers.
    """
    name = models.CharField(
        max_length=100,
        unique=True
    )
    slug = models.SlugField(
        max_length=100,
        unique=True
    )
    description = models.CharField(
        max_length=200,
        blank=True
    )
    contacts = GenericRelation(
        to='tenancy.ContactAssignment'
    )
    comments = models.TextField(
        blank=True
    )

    clone_fields = [
        'description', 'comments'
    ]

    class Meta:
        ordering = ['name']

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('plugins:netbox_inventory:supplier', args=[self.pk])


class Purchase(NetBoxModel):
    """
    Represents a purchase of a set of Assets from a Supplier.
    """
    name = models.CharField(
        max_length=100
    )
    supplier = models.ForeignKey(
        help_text='Legal entity this purchase was made at',
        to='netbox_inventory.Supplier',
        on_delete=models.PROTECT,
        related_name='purchases',
        blank=False,
        null=False,
    )
    date = models.DateField(
        help_text='Date when this purchase was made',
        blank=True,
        null=True,
    )
    description = models.CharField(
        max_length=200,
        blank=True
    )
    comments = models.TextField(
        blank=True
    )

    clone_fields = [
        'supplier', 'date', 'description', 'comments'
    ]

    class Meta:
        ordering = ['supplier', 'name']
        unique_together = (
            ('supplier', 'name'),
        )

    def __str__(self):
        return f'{self.supplier} {self.name}'

    def get_absolute_url(self):
        return reverse('plugins:netbox_inventory:purchase', args=[self.pk])


class InventoryItemType(NetBoxModel):
    """
    Inventory Item Type is a model (make, part number) of an Inventory Item. In
    that it is simmilar to Device Type or Module Type.
    """
    manufacturer = models.ForeignKey(
        to='dcim.Manufacturer',
        on_delete=models.PROTECT,
        related_name='inventoryitem_types'
    )
    model = models.CharField(
        max_length=100
    )
    slug = models.SlugField(
        max_length=100
    )
    part_number = models.CharField(
        max_length=50,
        blank=True,
        help_text='Discrete part number (optional)',
        verbose_name='Part Number',
    )
    inventoryitem_group = models.ForeignKey(
        to='netbox_inventory.InventoryItemGroup',
        on_delete=models.SET_NULL,
        related_name='inventoryitem_types',
        blank=True,
        null=True,
        verbose_name='Inventory Item Group',
    )
    comments = models.TextField(
        blank=True
    )
    images = GenericRelation(
        to='extras.ImageAttachment'
    )

    clone_fields = [
        'manufacturer',
    ]

    class Meta:
        ordering = ['manufacturer', 'model']
        unique_together = [
            ['manufacturer', 'model'],
            ['manufacturer', 'slug'],
        ]

    def __str__(self):
        return self.model

    def get_absolute_url(self):
        return reverse('plugins:netbox_inventory:inventoryitemtype', args=[self.pk])


class InventoryItemGroup(NestedGroupModel):
    """
    Inventory Item Groups are groups of simmilar InventoryItemTypes.
    This allows you to, for example, have one Group for all your 10G-LR SFP
    pluggables, from different manufacturers/with different part numbers.
    Inventory Item Groups can be nested.
    """
    slug = None # remove field that is defined on NestedGroupModel

    comments = models.TextField(
        blank=True
    )

    class Meta:
        ordering = ['name']
        constraints = (
            models.UniqueConstraint(
                fields=('parent', 'name'),
                name='%(app_label)s_%(class)s_parent_name'
            ),
            models.UniqueConstraint(
                fields=('name',),
                name='%(app_label)s_%(class)s_name',
                condition=models.Q(parent__isnull=True),
                violation_error_message="A top-level group with this name already exists."
            ),
        )

    def get_absolute_url(self):
        return reverse('plugins:netbox_inventory:inventoryitemgroup', args=[self.pk])



from netbox.views import generic
from utilities.utils import count_related
from .. import filtersets, forms, models, tables

__all__ = (
    'SupplierView',
    'SupplierListView',
    'SupplierEditView',
    'SupplierDeleteView',
    'SupplierBulkImportView',
    'SupplierBulkEditView',
    'SupplierBulkDeleteView',
)

class SupplierView(generic.ObjectView):
    queryset = models.Supplier.objects.all()

    def get_extra_context(self, request, instance):
        supplier_assets = models.Asset.objects.restrict(request.user, 'view').filter(
            purchase__supplier=instance
        )
        asset_table = tables.AssetTable(supplier_assets, user=request.user)
        asset_table.columns.hide('supplier')
        asset_table.configure(request)

        return {
            'asset_table': asset_table,
            'asset_count': models.Asset.objects.filter(purchase__supplier=instance).count(),
            'purchase_count': models.Purchase.objects.filter(supplier=instance).count(),
        }

class SupplierListView(generic.ObjectListView):
    queryset = models.Supplier.objects.annotate(
        purchase_count=count_related(models.Purchase, 'supplier'),
        asset_count=count_related(models.Asset, 'purchase__supplier'),
    )
    table = tables.SupplierTable
    filterset = filtersets.SupplierFilterSet
    filterset_form = forms.SupplierFilterForm


class SupplierEditView(generic.ObjectEditView):
    queryset = models.Supplier.objects.all()
    form = forms.SupplierForm


class SupplierDeleteView(generic.ObjectDeleteView):
    queryset = models.Supplier.objects.all()


class SupplierBulkImportView(generic.BulkImportView):
    queryset = models.Supplier.objects.all()   
    table = tables.SupplierTable 
    model_form = forms.SupplierImportForm


class SupplierBulkEditView(generic.BulkEditView):
    queryset = models.Supplier.objects.all()
    filterset = filtersets.SupplierFilterSet
    table = tables.SupplierTable
    form = forms.SupplierBulkEditForm


class SupplierBulkDeleteView(generic.BulkDeleteView):
    queryset = models.Supplier.objects.all()
    table = tables.SupplierTable



from utilities.testing import APIViewTestCases
from ..custom import APITestCase
from ...models import Supplier


class SupplierTest(
        APITestCase, 
        APIViewTestCases.GetObjectViewTestCase,
        APIViewTestCases.ListObjectsViewTestCase,
        APIViewTestCases.CreateObjectViewTestCase,
        APIViewTestCases.UpdateObjectViewTestCase,
        APIViewTestCases.DeleteObjectViewTestCase):
    model = Supplier
    brief_fields = ['display', 'id', 'name', 'slug', 'url']
    create_data = [
        {
            'name': 'Supplier 4',
            'slug': 'supplier4',
        },
        {
            'name': 'Supplier 5',
            'slug': 'supplier5',
        },
        {
            'name': 'Supplier 6',
            'slug': 'supplier6',
        },
    ]
    bulk_update_data = {
        'description': 'new description',
    }

    @classmethod
    def setUpTestData(cls) -> None:
        Supplier.objects.create(name='Supplier 1', slug='supplier1')
        Supplier.objects.create(name='Supplier 2', slug='supplier2')
        Supplier.objects.create(name='Supplier 3', slug='supplier3')
